@@ -34,14 +34,17 @@ import collections
 import functools
 import inspect
 import itertools
+import logging
 import os
 import random
 import re
 import shutil
 import string
+import subprocess
 import sys
 import time
 import unittest
+import weakref
 
 if sys.version_info[0] < 3:
 	import urllib
@@ -76,11 +79,11 @@ class BruteforceGenerator(object):
 		:param charset: The characters to include in the resulting sequences.
 		"""
 		self.startlen = startlen
-		if endlen == None:
+		if endlen is None:
 			self.endlen = startlen
 		else:
 			self.endlen = endlen
-		if charset == None:
+		if charset is None:
 			charset = list(map(chr, range(0, 256)))
 		elif isinstance(charset, str):
 			charset = list(charset)
@@ -126,13 +129,18 @@ class Cache(object):
 		self._target_function = None
 		self._target_function_arg_spec = None
 		self.__cache = {}
+		self.__obj = None
+
+	def __get__(self, instance, _):
+		self.__obj = instance
+		return self
 
 	def __call__(self, *args, **kwargs):
 		if not getattr(self, '_target_function', False):
 			target_function = args[0]
 			if not inspect.isfunction(target_function) and not inspect.ismethod(target_function):
 				raise RuntimeError('the cached object must be a function or method')
-			arg_spec = inspect.getargspec(target_function)
+			arg_spec = inspect.getargspec(target_function)  # pylint: disable=W1505
 			if arg_spec.varargs or arg_spec.keywords:
 				raise RuntimeError('the cached function can not use dynamic args or kwargs')
 			self._target_function = target_function
@@ -140,21 +148,34 @@ class Cache(object):
 			return functools.wraps(target_function)(self)
 
 		self.cache_clean()
+		if self.__obj is not None:
+			args = (self.__obj,) + args
+			self.__obj = None
+			is_method = True
+		else:
+			is_method = False
 		args = self._flatten_args(args, kwargs)
-		if not isinstance(args, collections.Hashable):
-			return self._target_function(*args)
-		result, expiration = self.__cache.get(args, (None, 0))
+		if is_method:
+			inst = args.popleft()
+			args = tuple(args)
+			ref = weakref.ref(inst, functools.partial(self._ref_callback, args))
+			cache_args = (ref,) + args
+			args = (inst,) + args
+		else:
+			cache_args = tuple(args)
+			args = tuple(args)
+		result, expiration = self.__cache.get(cache_args, (None, 0))
 		if expiration > time.time():
 			return result
 		result = self._target_function(*args)
-		self.__cache[args] = (result, time.time() + self.cache_timeout)
+		self.__cache[cache_args] = (result, time.time() + self.cache_timeout)
 		return result
 
 	def __repr__(self):
 		return "<cached function {0} at 0x{1:x}>".format(self._target_function.__name__, id(self._target_function))
 
 	def _flatten_args(self, args, kwargs):
-		flattened_args = list(args)
+		flattened_args = collections.deque(args)
 		arg_spec = self._target_function_arg_spec
 
 		arg_spec_defaults = (arg_spec.defaults or [])
@@ -166,25 +187,25 @@ class Cache(object):
 			if arg_name in default_args:
 				if not arg_name in kwargs:
 					raise TypeError("{0}() missing required argument '{1}'".format(self._target_function.__name__, arg_name))
-				flattened_args.append(kwargs[arg_name])
-				del kwargs[arg_name]
+				flattened_args.append(kwargs.pop(arg_name))
 			else:
-				flattened_args.append(kwargs.get(arg_name, default_kwargs[arg_name]))
-				if arg_name in kwargs:
-					del kwargs[arg_name]
+				flattened_args.append(kwargs.pop(arg_name, default_kwargs[arg_name]))
 
-		unexpected_kwargs = tuple(kwargs.keys())
-		if len(unexpected_kwargs):
-			unexpected_kwargs = tuple("'{0}'".format(a) for a in unexpected_kwargs)
+		if kwargs:
+			unexpected_kwargs = tuple("'{0}'".format(a) for a in kwargs.keys())
 			raise TypeError("{0}() got an unexpected keyword argument{1} {2}".format(self._target_function.__name__, ('' if len(unexpected_kwargs) == 1 else 's'), ', '.join(unexpected_kwargs)))
-		return tuple(flattened_args)
+		return flattened_args
+
+	def _ref_callback(self, args, ref):
+		args = (ref,) + args
+		self.__cache.pop(args, None)
 
 	def cache_clean(self):
 		"""
 		Remove expired items from the cache.
 		"""
 		now = time.time()
-		keys_for_removal = []
+		keys_for_removal = collections.deque()
 		for key, (_, expiration) in self.__cache.items():
 			if expiration < now:
 				keys_for_removal.append(key)
@@ -276,7 +297,7 @@ class SectionConfigParser(object):
 
 	def _get_raw(self, option, opt_type, default=None):
 		get_func = getattr(self.config_parser, 'get' + opt_type)
-		if default == None:
+		if default is None:
 			return get_func(self.section_name, option)
 		elif self.config_parser.has_option(self.section_name, option):
 			return get_func(self.section_name, option)
@@ -378,6 +399,42 @@ class TestCase(unittest.TestCase):
 		if not hasattr(self, 'assertRaisesRegex') and hasattr(self, 'assertRaisesRegexp'):
 			self.assertRaisesRegex = self.assertRaisesRegexp
 
+def configure_stream_logger(logger='', level=None, formatter='%(levelname)-8s %(message)s'):
+	"""
+	Configure the default stream handler for logging messages to the console,
+	remove other logging handlers, and enable capturing warnings.
+
+	.. versionadded:: 1.3.0
+
+	:param str logger: The logger to add the stream handler for.
+	:param level: The level to set the logger to, will default to WARNING if no level is specified.
+	:type level: None, int, str
+	:param formatter: The format to use for logging messages to the console.
+	:type formatter: str, :py:class:`logging.Formatter`
+	:return: The new configured stream handler.
+	:rtype: :py:class:`logging.StreamHandler`
+	"""
+	level = level or logging.WARNING
+	if isinstance(level, str):
+		level = getattr(logging, level, None)
+		if level is None:
+			raise ValueError('invalid log level: ' + level)
+	root_logger = logging.getLogger('')
+	for handler in root_logger.handlers:
+		root_logger.removeHandler(handler)
+
+	logging.getLogger(logger).setLevel(logging.DEBUG)
+	console_log_handler = logging.StreamHandler()
+	console_log_handler.setLevel(level)
+	if isinstance(formatter, str):
+		formatter = logging.Formatter(formatter)
+	elif not isinstance(formatter, logging.Formatter):
+		raise TypeError('formatter must be an instance of logging.Formatter')
+	console_log_handler.setFormatter(formatter)
+	logging.getLogger(logger).addHandler(console_log_handler)
+	logging.captureWarnings(True)
+	return console_log_handler
+
 def download(url, filename=None):
 	"""
 	Download a file from a url and save it to disk.
@@ -409,7 +466,7 @@ def escape_single_quote(unescaped):
 
 def format_bytes_size(val):
 	"""
-	Take a number of bytes and convert it to a human readble number.
+	Take a number of bytes and convert it to a human readable number.
 
 	:param int val: The number of bytes to format.
 	:return: The size in a human readable format.
@@ -455,6 +512,39 @@ def is_valid_email_address(email_address):
 	# requirements = re
 	return EMAIL_REGEX.match(email_address) != None
 
+def open_uri(uri):
+	"""
+	Open a URI in a platform intelligent way. On Windows this will use
+	'cmd.exe /c start' and on Linux this will use gvfs-open or xdg-open
+	depending on which is available. If no suitable application can be
+	found to open the URI, a RuntimeError will be raised.
+
+	.. versionadded:: 1.3.0
+
+	:param str uri: The URI to open.
+	"""
+	close_fds = True
+	startupinfo = None
+	proc_args = []
+	if sys.platform.startswith('win'):
+		proc_args.append(which('cmd.exe'))
+		proc_args.append('/c')
+		proc_args.append('start')
+		uri = uri.replace('&', '^&')
+		close_fds = False
+		startupinfo = subprocess.STARTUPINFO()
+		startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
+		startupinfo.wShowWindow = subprocess.SW_HIDE
+	elif which('gvfs-open'):
+		proc_args.append(which('gvfs-open'))
+	elif which('xdg-open'):
+		proc_args.append(which('xdg-open'))
+	else:
+		raise RuntimeError('could not find suitable application to open uri')
+	proc_args.append(uri)
+	proc_h = subprocess.Popen(proc_args, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE, close_fds=close_fds, startupinfo=startupinfo)
+	return proc_h.wait() == 0
+
 def parse_case_camel_to_snake(camel):
 	"""
 	Convert a string from CamelCase to snake_case.
@@ -509,13 +599,17 @@ def parse_server(server, default_port):
 def parse_timespan(timedef):
 	"""
 	Convert a string timespan definition to seconds, for example converting
-	'1m30s' to 90.
+	'1m30s' to 90. If *timedef* is already an int, the value will be returned
+	unmodified.
 
-	:param str timedef: The timespan definition to convert to seconds.
+	:param timedef: The timespan definition to convert to seconds.
+	:type timedef: int, str
 	:return: The converted value in seconds.
 	:rtype: int
 	"""
-	converter_order = ['w', 'd', 'h', 'm', 's']
+	if isinstance(timedef, int):
+		return timedef
+	converter_order = ('w', 'd', 'h', 'm', 's')
 	converters = {
 		'w': 604800,
 		'd': 86400,
@@ -560,10 +654,10 @@ def parse_to_slug(words, maxlen=24):
 	"""
 	slug = ''
 	maxlen = min(maxlen, len(words))
-	for i in range(len(words)):
+	for c in words:
 		if len(slug) == maxlen:
 			break
-		c = ord(words[i])
+		c = ord(c)
 		if c == 0x27:
 			continue
 		elif c >= 0x30 and c <= 0x39:
@@ -647,7 +741,7 @@ def unique(seq, key=None):
 	if key is None:
 		key = lambda x: x
 	preserved_type = type(seq)
-	if not preserved_type in (list, tuple):
+	if preserved_type not in (list, tuple):
 		raise TypeError("unique argument 1 must be list or tuple, not {0}".format(preserved_type.__name__))
 	seen = []
 	result = []
@@ -659,6 +753,37 @@ def unique(seq, key=None):
 		result.append(item)
 	return preserved_type(result)
 
+def weighted_choice(choices, weight):
+	"""
+	Make a random selection from the specified choices. Apply the *weight*
+	function to each to return a positive integer representing shares of
+	selection pool the choice should received. The *weight* function is passed a
+	single argument of the choice from the *choices* iterable.
+
+	:param choices: The choices to select from.
+	:type choices: list, tuple
+	:param weight: The function used for gather weight information for choices.
+	:type weight: function
+	:return: A randomly selected choice from the provided *choices*.
+	"""
+	# requirements = random
+	weights = []
+	# get weight values for each of the choices
+	for choice in choices:
+		choice_weight = weight(choice)
+		if not (isinstance(choice_weight, int) and choice_weight > 0):
+			raise TypeError('weight results must be positive integers')
+		weights.append(choice_weight)
+
+	# make a selection within the acceptable range
+	selection = random.randint(0, sum(weights) - 1)
+
+	# find and return the corresponding choice
+	for idx, choice in enumerate(choices):
+		if selection < sum(weights[:idx + 1]):
+			return choice
+	raise RuntimeError('no selection could be made')
+
 def which(program):
 	"""
 	Locate an executable binary's full path by its name.
@@ -689,9 +814,9 @@ def xfrange(start, stop=None, step=1):
 	:param step: Stepping size.
 	:type step: float, int, long
 	"""
-	if stop == None:
+	if stop is None:
 		stop = start
-		start = 0
+		start = 0.0
 	start = float(start)
 	while start < stop:
 		yield start
